/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package org.objectweb.speedo.stress;


import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.speedo.pobjects.common.ParamsTest;

import javax.jdo.PersistenceManager;
import javax.jdo.JDOFatalException;

/**
 * Stresses the deletion function of Speedo.
 * 
 * @author P. Dechamboux
 */
public class TestDeletion extends IntUserIdHelper {

	private String DELETION = getLoggerName() + ".deletion";


	public TestDeletion(String s) {
		super(s);
	}

	protected String getLoggerName() {
		return STRESS_LOG_NAME + ".TestDeletion";
	}

	class DeletionCtx extends IUICtx {
		public int nbDeletion;

		public DeletionCtx(int dbSize, int nbDeletion) {
			super(dbSize);
			this.nbDeletion = nbDeletion;
		}

		public String toString() {
			return super.toString() + " / ndDeletion = " + nbDeletion;
		}
	}

	protected void prepareTest(StressHelper.TaskManager tm, Object ctx) {
		oids = null;
		super.prepareTest(tm, ctx);
	}

	protected void perform(Task task, int threadId, int txId, Object ctx, PerformResult res) {
		DeletionCtx dctx = (DeletionCtx) ctx;
		PersistenceManager pm = getPM(task, threadId,txId);
		try {
			res.beginTest();
			beginTx(pm, task, threadId, txId);
			for (int no = 0; no < dctx.nbDeletion; no++) {
				int oid = (txId * dctx.nbDeletion) + no;
				if (debug) {
					logger.log(BasicLevel.DEBUG, "Deleting object "  + oid);
				}
				pm.deletePersistent(pm.getObjectById(dctx.oids[oid], false));
			}
			commitTx(pm, task, threadId, txId);
			res.endTest();
		} catch (JDOFatalException e) {
			rollbackOnException(pm, e, res, task, threadId, txId);
		} catch (Throwable e) {
			stopOnError(pm, e, res, task, threadId, txId);
		} finally {
			closePM(pm, threadId, txId, task, res);
		}
	}


	/**
	 * Methode d'aiguillage
	 */
	protected void perform(ParamsTest param) {
		if (!interactive) {
			perform(param.getNbth(), param.getNbtx(), param.getTo(),
			        new DeletionCtx(param.getDbSize(), param.getNbc()));
		}
	}

	/**
	 * Tests the deletion of a lot of persistent objects, with interactive
	 * setting of test parameteres (see file userconf/project.properties).
	 */
	public void testInteractive() {
		if (interactive) {
			perform(Integer.getInteger(THREAD, 10).intValue(),
			        Integer.getInteger(TX, 100).intValue(),
			        Integer.getInteger(TIMEOUT, 200000).intValue(),
			        new DeletionCtx(
			            Integer.getInteger(DBSIZE, 100000).intValue(),
			            Integer.getInteger(DELETION, 1000).intValue()));
		}
	}

	/**
	 * Tests 10 transactions where each of them creates 1 object using 1
	 * thread.
	 */
	public void testDeletionTh1Tx10Cr1() {
		if (!interactive) {
			perform(1, 10, 1000000, new DeletionCtx(10, 1));
		}
	}
/* *
}
class Toto {
	boolean interactive;
	void perform(ParamsTest p){}
* */

	/**
	 * Tests 100 transactions where each of them Delete 1 object using 1
	 * thread.
	 */
	public void testDeletionTh1Tx100De1() {
		ParamsTest params = new ParamsTest(1, 100, 1, 1000000, 0, 100);
		if (!interactive) {
			perform(params);
		}
	}
	/**
	 * Tests 1.000 transactions where each of them deletes 1 object using 1
	 * thread.
	 */
	public void testDeletionTh1Tx1000Cr1() {
		ParamsTest params = new ParamsTest(1, 1000, 1, 1000000, 0, 1000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10.000 transactions where each of them deletes 1 object using 1
	 * thread.
	 */
	public void testDeletionTh1Tx10000Cr1() {
		ParamsTest params = new ParamsTest(1, 10000, 1, 1000000, 0, 10000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100.000 transactions where each of them deletes 1 object using 1
	 * thread.
	 */
	public void testDeletionTh1Tx100000Cr1() {


		ParamsTest params = new ParamsTest(1, 100000, 1, 10000000, 0, 100000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10 transactions where each of them deletes 10 objects using 1
	 * thread.
	 */
	public void testDeletionTh1Tx10Cr10() {


		ParamsTest params = new ParamsTest(1, 10, 10, 1000000, 0, 100);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100 transactions where each of them deletes 10 objects using 1
	 * thread.
	 */
	public void testDeletionTh1Tx100Cr10() {


		ParamsTest params = new ParamsTest(1, 100, 10, 1000000, 0, 1000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 1.000 transactions where each of them deletes 10 objects using 1
	 * thread.
	 */
	public void testDeletionTh1Tx1000Cr10() {


		ParamsTest params = new ParamsTest(1, 1000, 10, 1000000, 0, 10000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10.000 transactions where each of them deletes 10 objects using 1
	 * thread.
	 */
	public void testDeletionTh1Tx10000Cr10() {


		ParamsTest params = new ParamsTest(1, 10000, 10, 10000000, 0, 100000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100.000 transactions where each of them deletes 10 objects using 1
	 * thread.
	 */
	public void _testDeletionTh1Tx100000Cr10() {


		ParamsTest params = new ParamsTest(1, 100000, 10, 100000000, 0, 1000000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10 transactions where each of them deletes 100 objects using 1
	 * thread.
	 */
	public void testDeletionTh1Tx10Cr100() {


		ParamsTest params = new ParamsTest(1, 10, 100, 1000000, 0, 1000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100 transactions where each of them deletes 100 objects using 1
	 * thread.
	 */
	public void testDeletionTh1Tx100Cr100() {


		ParamsTest params = new ParamsTest(1, 100, 100, 1000000, 0, 10000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 1.000 transactions where each of them deletes 100 objects using 1
	 * thread.
	 */
	public void testDeletionTh1Tx1000Cr100() {


		ParamsTest params = new ParamsTest(1, 1000, 100, 10000000, 0, 100000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10.000 transactions where each of them deletes 100 objects using 1
	 * thread.
	 */
	public void testDeletionTh1Tx10000Cr100() {


		ParamsTest params = new ParamsTest(1, 10000, 100, 100000000, 0, 1000000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10 transactions where each of them deletes 500 objects using 1
	 * thread.
	 */
	public void testDeletionTh1Tx10Cr500() {


		ParamsTest params = new ParamsTest(1, 10, 500, 1000000, 0, 5000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100 transactions where each of them deletes 500 objects using 1
	 * thread.
	 */
	public void testDeletionTh1Tx100Cr500() {


		ParamsTest params = new ParamsTest(1, 100, 500, 10000000, 0, 50000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 1.000 transactions where each of them deletes 500 objects using 1
	 * thread.
	 */
	public void testDeletionTh1Tx1000Cr500() {


		ParamsTest params = new ParamsTest(1, 1000, 500, 100000000, 0, 500000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10 transactions where each of them deletes 1.000 objects using 1
	 * thread.
	 */
	public void testDeletionTh1Tx10Cr1000() {


		ParamsTest params = new ParamsTest(1, 10, 1000, 1000000, 0, 10000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100 transactions where each of them deletes 1.000 objects using 1
	 * thread.
	 */
	public void testDeletionTh1Tx100Cr1000() {


		ParamsTest params = new ParamsTest(1, 100, 1000, 10000000, 0, 100000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 1.000 transactions where each of them deletes 1.000 objects using
	 * 1 thread.
	 */
	public void testDeletionTh1Tx1000Cr1000() {


		ParamsTest params = new ParamsTest(1, 1000, 1000, 100000000, 0, 1000000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10 transactions where each of them deletes 2.000 objects using 1
	 * thread.
	 */
	public void testDeletionTh1Tx10Cr2000() {


		ParamsTest params = new ParamsTest(1, 10, 2000, 10000000, 0, 20000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100 transactions where each of them deletes 2.000 objects using 1
	 * thread.
	 */
	public void testDeletionTh1Tx100Cr2000() {


		ParamsTest params = new ParamsTest(1, 100, 2000, 100000000, 0, 200000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10 transactions where each of them deletes 1 object using 2
	 * threads.
	 */
	public void testDeletionTh2Tx10Cr1() {


		ParamsTest params = new ParamsTest(2, 10, 1, 1000000, 0, 10);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100 transactions where each of them deletes 1 object using 2
	 * threads.
	 */
	public void testDeletionTh2Tx100Cr1() {


		ParamsTest params = new ParamsTest(2, 100, 1, 1000000, 0, 100);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 1.000 transactions where each of them deletes 1 object using 2
	 * threads.
	 */
	public void testDeletionTh2Tx1000Cr1() {


		ParamsTest params = new ParamsTest(2, 1000, 1, 1000000, 0, 1000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10.000 transactions where each of them deletes 1 object using 2
	 * threads.
	 */
	public void testDeletionTh2Tx10000Cr1() {


		ParamsTest params = new ParamsTest(2, 10000, 1, 1000000, 0, 10000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100.000 transactions where each of them deletes 1 object using 2
	 * threads.
	 */
	public void _testDeletionTh2Tx100000Cr1() {


		ParamsTest params = new ParamsTest(2, 100000, 1, 1000000, 0, 100000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10 transactions where each of them deletes 10 objects using 2
	 * threads.
	 */
	public void testDeletionTh2Tx10Cr10() {


		ParamsTest params = new ParamsTest(2, 10, 10, 1000000, 0, 100);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100 transactions where each of them deletes 10 objects using 2
	 * threads.
	 */
	public void testDeletionTh2Tx100Cr10() {


		ParamsTest params = new ParamsTest(2, 100, 10, 1000000, 0, 1000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 1.000 transactions where each of them deletes 10 objects using 2
	 * threads.
	 */
	public void testDeletionTh2Tx1000Cr10() {


		ParamsTest params = new ParamsTest(2, 1000, 10, 1000000, 0, 10000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10.000 transactions where each of them deletes 10 objects using 2
	 * threads.
	 */
	public void testDeletionTh2Tx10000Cr10() {


		ParamsTest params = new ParamsTest(2, 10000, 10, 1000000, 0, 100000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100.000 transactions where each of them deletes 10 objects using 2
	 * threads.
	 */
	public void _testDeletionTh2Tx100000Cr10() {


		ParamsTest params = new ParamsTest(2, 100000, 10, 1000000, 0, 1000000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10 transactions where each of them deletes 100 objects using 2
	 * threads.
	 */
	public void testDeletionTh2Tx10Cr100() {


		ParamsTest params = new ParamsTest(2, 10, 100, 1000000, 0, 1000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100 transactions where each of them deletes 100 objects using 2
	 * threads.
	 */
	public void testDeletionTh2Tx100Cr100() {


		ParamsTest params = new ParamsTest(2, 100, 100, 1000000, 0, 10000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 1.000 transactions where each of them deletes 100 objects using 2
	 * threads.
	 */
	public void testDeletionTh2Tx1000Cr100() {


		ParamsTest params = new ParamsTest(2, 1000, 100, 1000000, 0, 100000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100 transactions where each of them deletes 2.000 objects using 2
	 * threads.
	 */
	public void testDeletionTh2Tx100Cr2000() {


		ParamsTest params = new ParamsTest(2, 100, 2000, 10000000, 0, 200000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 1.000 transactions where each of them deletes 2.000 objects using
	 * 2 threads.
	 */
	public void testDeletionTh2Tx1000Cr2000() {


		ParamsTest params = new ParamsTest(2, 1000, 2000, 10000000, 0, 2000000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10 transactions where each of them deletes 1 object using 3
	 * threads.
	 */
	public void testDeletionTh3Tx10Cr1() {


		ParamsTest params = new ParamsTest(3, 10, 1, 1000000, 0, 10);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100 transactions where each of them deletes 1 object using 3
	 * threads.
	 */
	public void testDeletionTh3Tx100Cr1() {


		ParamsTest params = new ParamsTest(3, 100, 1, 1000000, 0, 100);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 1.000 transactions where each of them deletes 1 object using 3
	 * threads.
	 */
	public void testDeletionTh3Tx1000Cr1() {


		ParamsTest params = new ParamsTest(3, 1000, 1, 1000000, 0, 1000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10.000 transactions where each of them deletes 1 object using 3
	 * threads.
	 */
	public void testDeletionTh3Tx10000Cr1() {


		ParamsTest params = new ParamsTest(3, 10000, 1, 1000000, 0, 10000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100.000 transactions where each of them deletes 1 object using 3
	 * threads.
	 */
	public void _testDeletionTh3Tx100000Cr1() {


		ParamsTest params = new ParamsTest(3, 100000, 1, 1000000, 0, 100000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10 transactions where each of them deletes 10 objects using 3
	 * threads.
	 */
	public void testDeletionTh3Tx10Cr10() {


		ParamsTest params = new ParamsTest(3, 10, 10, 1000000, 0, 100);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100 transactions where each of them deletes 10 objects using 3
	 * threads.
	 */
	public void testDeletionTh3Tx100Cr10() {


		ParamsTest params = new ParamsTest(3, 100, 10, 1000000, 0, 1000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 1.000 transactions where each of them deletes 10 objects using 3
	 * threads.
	 */
	public void testDeletionTh3Tx1000Cr10() {


		ParamsTest params = new ParamsTest(3, 1000, 10, 1000000, 0, 10000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10.000 transactions where each of them deletes 10 objects using 3
	 * threads.
	 */
	public void testDeletionTh3Tx10000Cr10() {


		ParamsTest params = new ParamsTest(3, 10000, 10, 1000000, 0, 100000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100.000 transactions where each of them deletes 10 objects using 3
	 * threads.
	 */
	public void _testDeletionTh3Tx100000Cr10() {


		ParamsTest params = new ParamsTest(3, 100000, 10, 1000000, 0, 1000000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100 transactions where each of them deletes 1 object using 4
	 * threads.
	 */
	public void testDeletionTh4Tx100Cr1() {


		ParamsTest params = new ParamsTest(4, 100, 1, 1000000, 0, 100);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 1.000 transactions where each of them deletes 1 object using 4
	 * threads.
	 */
	public void testDeletionTh4Tx1000Cr1() {


		ParamsTest params = new ParamsTest(4, 1000, 1, 1000000, 0, 1000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10.000 transactions where each of them deletes 1 object using 4
	 * threads.
	 */
	public void testDeletionTh4Tx10000Cr1() {


		ParamsTest params = new ParamsTest(4, 10000, 1, 1000000, 0, 10000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100.000 transactions where each of them deletes 1 object using 4
	 * threads.
	 */
	public void _testDeletionTh4Tx100000Cr1() {



		ParamsTest params = new ParamsTest(4, 100000, 1, 1000000, 0, 100000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10 transactions where each of them deletes 10 objects using 4
	 * threads.
	 */
	public void testDeletionTh4Tx10Cr10() {


		ParamsTest params = new ParamsTest(4, 10, 10, 1000000, 0, 100);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 1.000 transactions where each of them deletes 10 objects using 4
	 * threads.
	 */
	public void testDeletionTh4Tx1000Cr10() {


		ParamsTest params = new ParamsTest(4, 1000, 10, 1000000, 0, 10000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10.000 transactions where each of them deletes 10 objects using 4
	 * threads.
	 */
	public void testDeletionTh4Tx10000Cr10() {


		ParamsTest params = new ParamsTest(4, 10000, 10, 1000000, 0, 100000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100.000 transactions where each of them deletes 10 objects using 4
	 * threads.
	 */
	public void _testDeletionTh4Tx100000Cr10() {


		ParamsTest params = new ParamsTest(4, 100000, 10, 1000000, 0, 1000000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100 transactions where each of them deletes 500 objects using 4
	 * threads.
	 */
	public void testDeletionTh4Tx100Cr500() {


		ParamsTest params = new ParamsTest(4, 100, 500, 1000000, 0, 50000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 1.000 transactions where each of them deletes 500 objects using 4
	 * threads.
	 */
	public void testDeletionTh4Tx1000Cr500() {


		ParamsTest params = new ParamsTest(4, 1000, 500, 1000000, 0, 500000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100 transactions where each of them deletes 10.000 objects using 4
	 * threads.
	 */
	public void testDeletionTh4Tx100Cr10000() {


		ParamsTest params = new ParamsTest(4, 100, 10000, 10000000, 1000000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10 transactions where each of them deletes 1 object using 10
	 * threads.
	 */
	public void testDeletionTh10Tx10Cr1() {


		ParamsTest params = new ParamsTest(10, 10, 1, 1000000, 0, 10);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100 transactions where each of them deletes 1 object using 10
	 * threads.
	 */
	public void testDeletionTh10Tx100Cr1() {


		ParamsTest params = new ParamsTest(10, 100, 1, 1000000, 0, 100);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 1.000 transactions where each of them deletes 1 object using 10
	 * threads.
	 */
	public void testDeletionTh10Tx1000Cr1() {


		ParamsTest params = new ParamsTest(10, 1000, 1, 1000000, 0, 1000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10.000 transactions where each of them deletes 1 object using 10
	 * threads.
	 */
	public void testDeletionTh10Tx10000Cr1() {


		ParamsTest params = new ParamsTest(10, 10000, 1, 1000000, 0, 10000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100.000 transactions where each of them deletes 1 object using 10
	 * threads.
	 */
	public void _testDeletionTh10Tx100000Cr1() {


		ParamsTest params = new ParamsTest(10, 100000, 1, 1000000, 0, 100000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10 transactions where each of them deletes 10 objects using 10
	 * threads.
	 */
	public void testDeletionTh10Tx10Cr10() {


		ParamsTest params = new ParamsTest(10, 10, 10, 1000000, 0, 100);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100 transactions where each of them deletes 10 objects using 10
	 * threads.
	 */
	public void testDeletionTh10Tx100Cr10() {


		ParamsTest params = new ParamsTest(10, 100, 10, 1000000, 0, 1000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 1.000 transactions where each of them deletes 10 objects using 10
	 * threads.
	 */
	public void testDeletionTh10Tx1000Cr10() {


		ParamsTest params = new ParamsTest(10, 1000, 10, 1000000, 0, 10000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 10.000 transactions where each of them deletes 10 objects using 10
	 * threads.
	 */
	public void testDeletionTh10Tx10000Cr10() {


		ParamsTest params = new ParamsTest(10, 10000, 10, 1000000, 0, 100000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100.000 transactions where each of them deletes 10 objects using
	 * 10 threads.
	 */
	public void _testDeletionTh10Tx100000Cr10() {


		ParamsTest params = new ParamsTest(10, 100000, 10, 1000000, 0, 1000000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100 transactions where each of them deletes 500 objects using 10
	 * threads.
	 */
	public void testDeletionTh10Tx100Cr500() {


		ParamsTest params = new ParamsTest(10, 100, 500, 1000000, 0, 50000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 1.000 transactions where each of them deletes 500 objects using 10
	 * threads.
	 */
	public void testDeletionTh10Tx1000Cr500() {


		ParamsTest params = new ParamsTest(10, 1000, 500, 1000000, 0, 500000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 1.000 transactions where each of them deletes 1.000 objects using
	 * 10 threads.
	 */
	public void testDeletionTh10Tx1000Cr1000() {


		ParamsTest params = new ParamsTest(10, 1000, 1000, 1000000, 0, 1000000);
		if (!interactive) {
			perform(params);
		}
	}

	/**
	 * Tests 100 transactions where each of them deletes 10.000 objects using
	 * 10 threads.
	 */
	public void testDeletionTh10Tx100Cr10000() {



		ParamsTest params = new ParamsTest(10, 100, 10000, 10000000, 0, 1000000);

		if (!interactive) {
			perform(params);
		}
	}

}
